from datetime import datetime, timedelta
from unittest.mock import MagicMock

import pytest

from sklik.api import SklikApi
from sklik.object import Account
from sklik.report import (
    _restriction_filter_validator,
    _display_option_validator,
    Report,
    create_report,
)
from sklik.util import SKLIK_DATE_FORMAT


# ========= Tests for helper functions =========

def test_restriction_filter_validator_defaults():
    # When since and until are None, defaults should be used.
    result = _restriction_filter_validator(None, None, None)
    expected_since = (datetime.today() - timedelta(days=29)).strftime(SKLIK_DATE_FORMAT)
    expected_until = (datetime.today() - timedelta(days=1)).strftime(SKLIK_DATE_FORMAT)
    assert result["dateFrom"] == expected_since
    assert result["dateTo"] == expected_until


def test_restriction_filter_validator_with_values():
    # When since and until are provided, they should be used.
    result = _restriction_filter_validator("2025-01-01", "2025-01-31", {"someKey": "someValue"})
    assert result["dateFrom"] == "2025-01-01"
    assert result["dateTo"] == "2025-01-31"
    assert result["someKey"] == "someValue"


def test_restriction_filter_validator_preserve_existing_keys():
    # If the restriction_filter already contains a date key, setdefault will not override it.
    result = _restriction_filter_validator(
        "2025-01-01", "2025-01-31", {"dateFrom": "existing", "other": "value"}
    )
    assert result["dateFrom"] == "existing"
    # dateTo is missing so it should be set from the argument.
    assert result["dateTo"] == "2025-01-31"
    assert result["other"] == "value"


def test_display_option_validator_default():
    # When granularity is None, the default value should be returned.
    result = _display_option_validator(None)
    assert result == {"statGranularity": "daily"}


def test_display_option_validator_with_value():
    # When a granularity is provided, it should be used.
    result = _display_option_validator("weekly")
    assert result == {"statGranularity": "weekly"}


# ========= Fixtures for Report tests =========

@pytest.fixture
def fake_api():
    """A fake SklikApi with a mocked call method."""
    return MagicMock(spec=SklikApi)


# ========= Tests for the Report class =========

def test_load_page(fake_api):
    # Simulate a raw API response for the 'readReport' call.
    raw_report = [
        {
            "someField": "a",
            "stats": [
                {"field1": "value1"},
                {"field1": "value2"},
            ],
        }
    ]
    # Configure the fake API so that any call to readReport returns our fake data.
    fake_api.call.return_value = {"report": raw_report}

    # Create a Report instance. Its __init__ will call load_page immediately.
    report = Report(
        account_id=1,
        report_id="rep123",
        service="reportService",
        fields=["field1"],
        total_count=2,
        api=fake_api,
    )

    # Expected output after merging each 'stat' with the remaining keys.
    expected_output = [
        {"field1": "value1", "someField": "a"},
        {"field1": "value2", "someField": "a"},
    ]

    # The initial page loaded in __init__ should match our expected output.
    assert report._current_page == expected_output

    # Also, test the load_page method directly.
    output = report.load_page(0, 100)
    assert output == expected_output

    # Verify that the API was called with the expected arguments.
    fake_api.call.assert_called_with(
        "reportService",
        "readReport",
        args=[
            {"userId": 1},
            "rep123",
            {"offset": 0, "limit": 100, "displayColumns": ["field1"]},
        ],
    )


def test_report_iteration(fake_api):
    # Configure the fake API to return one page with data, then an empty page.
    raw_report_page1 = [
        {
            "someField": "a",
            "stats": [
                {"field1": "value1"},
            ],
        }
    ]
    raw_report_page2 = []  # Simulate no more data.

    # The first call (in __init__) returns page1; the next call (during iteration) returns page2.
    fake_api.call.side_effect = [
        {"report": raw_report_page1},  # For __init__
        {"report": raw_report_page2},  # For the next page load
    ]

    report = Report(
        account_id=1,
        report_id="rep123",
        service="reportService",
        fields=["field1"],
        total_count=1,
        api=fake_api,
    )

    # Iteration should yield only the items from page1.
    items = list(report)
    expected_items = [{"field1": "value1", "someField": "a"}]
    assert items == expected_items


def test_report_repr_and_len(fake_api):
    # Use an API response that yields an empty page.
    fake_api.call.return_value = {"report": []}

    report = Report(
        account_id=42,
        report_id="rep456",
        service="service",
        fields=["field1"],
        total_count=10,
        api=fake_api,
    )

    rep_str = repr(report)
    assert "account_id=42" in rep_str
    assert "report_id=rep456" in rep_str
    assert "total_count=10" in rep_str
    assert len(report) == 10


# ========= Test for create_report =========

def test_create_report(fake_api):
    # Create a fake Account object with a mocked call method.
    fake_account = MagicMock(spec=Account)
    fake_account.account_id = 1
    fake_account.api = fake_api

    # Configure the account.call (used for 'createReport') to return a specific response.
    fake_create_response = {
        "reportId": "report_789",
        "totalCount": 5,
    }
    fake_account.call.return_value = fake_create_response

    # Also, ensure that Report.__init__ (which calls load_page) gets an empty report.
    fake_api.call.return_value = {"report": []}

    # Create a report using the create_report helper.
    report = create_report(
        account=fake_account,
        service="reportService",
        fields=["field1"],
        since="2025-01-01",
        until="2025-01-31",
        granularity="weekly",
        restriction_filter={"custom": "filter"},
    )

    # Build the expected arguments for the createReport call.
    expected_args = [
        _restriction_filter_validator("2025-01-01", "2025-01-31", {"custom": "filter"}),
        _display_option_validator("weekly"),
    ]
    fake_account.call.assert_called_with("reportService", "createReport", expected_args)

    # Verify that the returned Report has the attributes from the fake_create_response.
    assert report.account_id == 1
    assert report.report_id == "report_789"
    assert report._total_count == 5
    assert report._api == fake_account.api
